class O:
    def __init__(self, n: int, *args, **kwargs) -> None:
        """
        Initializes a `O` instance.

        Args:
        -----
        n : int
            The dimension of the O_n bundle.
        """
        raise NotImplementedError("This class is not implemented yet.")
